@@ -24,7 +24,13 @@
 #include <taler/taler_json_lib.h>
 #include <regex.h>
 #include "anastasis_util_lib.h"
+#include <gnunet/gnunet_db_lib.h>
+#include "anastasis_database_lib.h"
 
+/**
+ * How many retries do we allow per code?
+ */
+#define INITIAL_RETRY_COUNTER 3
 
 /**
  * Saves the State of a authorization plugin.
@@ -48,6 +54,11 @@ struct Email_Context
    */
   json_t *messages;
 
+  /**
+   * Context we operate in.
+   */
+  const struct ANASTASIS_AuthorizationContext *ac;
+
 };
 
 
@@ -160,8 +171,10 @@ get_message (const json_t *messages,
                            NULL, NULL))
     {
       GNUNET_break (0);
+      GNUNET_JSON_parse_free (spec);
       return NULL;
     }
+    GNUNET_JSON_parse_free (spec);
     return ret;
   }
 }
@@ -243,7 +256,20 @@ email_start (void *cls,
 {
   struct Email_Context *ctx = cls;
   struct ANASTASIS_AUTHORIZATION_State *as;
+  enum GNUNET_DB_QueryStatus qs;
 
+  /* If the user can show this challenge code, this
+     plugin is already happy (no additional
+     requirements), so mark this challenge as
+     already satisfied from the start. */
+  qs = ctx->ac->db->mark_challenge_code_satisfied (ctx->ac->db->cls,
+                                                   truth_uuid,
+                                                   code);
+  if (qs <= 0)
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
   as = GNUNET_new (struct ANASTASIS_AUTHORIZATION_State);
   as->trigger = trigger;
   as->trigger_cls = trigger_cls;
@@ -284,17 +310,20 @@ email_done_cb (void *cls,
  * I.e. start to send SMS or e-mail or launch video identification.
  *
  * @param as authorization state
+ * @param timeout how long do we have to produce a reply
  * @param connection HTTP client request (for queuing response, such as redirection to video portal)
  * @return state of the request
  */
 static enum ANASTASIS_AUTHORIZATION_Result
 email_process (struct ANASTASIS_AUTHORIZATION_State *as,
+               struct GNUNET_TIME_Absolute timeout,
                struct MHD_Connection *connection)
 {
   MHD_RESULT mres;
   const char *mime;
   const char *lang;
 
+  (void) timeout;
   mime = MHD_lookup_connection_value (connection,
                                       MHD_HEADER_KIND,
                                       MHD_HTTP_HEADER_ACCEPT);
@@ -519,18 +548,23 @@ email_cleanup (struct ANASTASIS_AUTHORIZATION_State *as)
 void *
 libanastasis_plugin_authorization_email_init (void *cls)
 {
+  const struct ANASTASIS_AuthorizationContext *ac = cls;
   struct ANASTASIS_AuthorizationPlugin *plugin;
-  struct GNUNET_CONFIGURATION_Handle *cfg = cls;
+  const struct GNUNET_CONFIGURATION_Handle *cfg = ac->cfg;
   struct Email_Context *ctx;
 
   ctx = GNUNET_new (struct Email_Context);
+  ctx->ac = ac;
   {
     char *fn;
     json_error_t err;
+    char *tmp;
 
+    tmp = GNUNET_OS_installation_get_path (GNUNET_OS_IPK_DATADIR);
     GNUNET_asprintf (&fn,
                      "%sauthorization-email-messages.json",
-                     GNUNET_OS_installation_get_path (GNUNET_OS_IPK_DATADIR));
+                     tmp);
+    GNUNET_free (tmp);
     ctx->messages = json_load_file (fn,
                                     JSON_REJECT_DUPLICATES,
                                     &err);
@@ -565,6 +599,7 @@ libanastasis_plugin_authorization_email_init (void *cls)
   }
 
   plugin = GNUNET_new (struct ANASTASIS_AuthorizationPlugin);
+  plugin->retry_counter = INITIAL_RETRY_COUNTER;
   plugin->code_validity_period = GNUNET_TIME_UNIT_DAYS;
   plugin->code_rotation_period = GNUNET_TIME_UNIT_HOURS;
   plugin->code_retransmission_frequency = GNUNET_TIME_UNIT_MINUTES;
