@@ -111,6 +111,11 @@ static int AH_connection_close;
  */
 static struct GNUNET_SCHEDULER_Task *mhd_task;
 
+/**
+ * Heap for processing timeouts of requests.
+ */
+struct GNUNET_CONTAINER_Heap *AH_to_heap;
+
 /**
  * Global return code
  */
@@ -332,6 +337,7 @@ url_handler (void *cls,
     hc = GNUNET_new (struct TM_HandlerContext);
     *con_cls = hc;
     hc->async_scope_id = aid;
+    hc->url = url;
   }
   if (0 == strcasecmp (method,
                        MHD_HTTP_METHOD_HEAD))
@@ -495,6 +501,11 @@ do_shutdown (void *cls)
     ANASTASIS_DB_plugin_unload (db);
     db = NULL;
   }
+  if (NULL != AH_to_heap)
+  {
+    GNUNET_CONTAINER_heap_destroy (AH_to_heap);
+    AH_to_heap = NULL;
+  }
 }
 
 
@@ -519,14 +530,36 @@ handle_mhd_completion_callback (void *cls,
                                 enum MHD_RequestTerminationCode toe)
 {
   struct TM_HandlerContext *hc = *con_cls;
+  struct GNUNET_AsyncScopeSave old_scope;
 
   (void) cls;
   (void) connection;
   if (NULL == hc)
     return;
-  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-              "Finished handling request with status %d\n",
-              (int) toe);
+  GNUNET_async_scope_enter (&hc->async_scope_id,
+                            &old_scope);
+  {
+#if MHD_VERSION >= 0x00097304
+    const union MHD_ConnectionInfo *ci;
+    unsigned int http_status = 0;
+
+    ci = MHD_get_connection_info (connection,
+                                  MHD_CONNECTION_INFO_HTTP_STATUS);
+    if (NULL != ci)
+      http_status = ci->http_status;
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Request for `%s' completed with HTTP status %u (%d)\n",
+                hc->url,
+                http_status,
+                toe);
+#else
+    (void) connection;
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Request for `%s' completed (%d)\n",
+                hc->url,
+                toe);
+#endif
+  }
   if (NULL != hc->cc)
     hc->cc (hc);
   GNUNET_free (hc);
@@ -708,12 +741,12 @@ run (void *cls,
   }
   if (GNUNET_OK !=
       GNUNET_CONFIGURATION_get_value_string (config,
-                                             "anastasis",
+                                             "anastasis-merchant-backend",
                                              "PAYMENT_BACKEND_URL",
                                              &AH_backend_url))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
-                               "anastasis",
+                               "anastasis-merchant-backend",
                                "PAYMENT_BACKEND_URL");
     GNUNET_SCHEDULER_shutdown ();
     return;
@@ -726,7 +759,7 @@ run (void *cls,
                           strlen ("http://"))) )
   {
     GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
-                               "anastasis",
+                               "anastasis-merchant-backend",
                                "PAYMENT_BACKEND_URL",
                                "Must be HTTP(S) URL");
     GNUNET_SCHEDULER_shutdown ();
@@ -739,7 +772,7 @@ run (void *cls,
                          AH_backend_url)) )
   {
     GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
-                               "anastasis",
+                               "anastasis-merchant-backend",
                                "PAYMENT_BACKEND_URL",
                                "Must have domain name");
     GNUNET_SCHEDULER_shutdown ();
@@ -823,6 +856,13 @@ run (void *cls,
                              certfile,
                              keyfile,
                              keypass);
+  if (NULL == apikey)
+  {
+    (void) GNUNET_CONFIGURATION_get_value_string (config,
+                                                  "anastasis-merchant-backend",
+                                                  "API_KEY",
+                                                  &apikey);
+  }
   if (NULL != apikey)
   {
     char *auth_header;
@@ -848,6 +888,14 @@ run (void *cls,
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
+  if (GNUNET_OK !=
+      db->connect (db->cls))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Database not setup. Did you run anastasis-dbinit?\n");
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
 
   fh = TALER_MHD_bind (config,
                        "anastasis",
