@@ -56,6 +56,7 @@
  */
 #define INITIAL_RETRY_COUNTER 3
 
+
 struct GetContext
 {
 
@@ -75,7 +76,7 @@ struct GetContext
   struct ANASTASIS_CRYPTO_TruthKeyP truth_key;
 
   /**
-   * true if client provided a payment secret / order ID?
+   * Cost for paying the challenge.
    */
   struct TALER_Amount challenge_cost;
 
@@ -125,7 +126,18 @@ struct GetContext
   struct MHD_Response *resp;
 
   /**
-   * How long do we wait at most for payment?
+   * Our entry in the #to_heap, or NULL.
+   */
+  struct GNUNET_CONTAINER_HeapNode *hn;
+
+  /**
+   * Challenge response we got from the request.
+   */
+  struct GNUNET_HashCode challenge_response;
+
+  /**
+   * How long do we wait at most for payment or
+   * authorization?
    */
   struct GNUNET_TIME_Absolute timeout;
 
@@ -218,6 +230,50 @@ static struct GetContext *gc_head;
  */
 static struct GetContext *gc_tail;
 
+/**
+ * Task running #do_timeout().
+ */
+static struct GNUNET_SCHEDULER_Task *to_task;
+
+
+/**
+ * Timeout requests that are past their due date.
+ *
+ * @param cls NULL
+ */
+static void
+do_timeout (void *cls)
+{
+  struct GetContext *gc;
+
+  (void) cls;
+  to_task = NULL;
+  while (NULL !=
+         (gc = GNUNET_CONTAINER_heap_peek (AH_to_heap)))
+  {
+    if (GNUNET_TIME_absolute_is_future (gc->timeout))
+      break;
+    if (gc->suspended)
+    {
+      /* Test needed as we may have a "concurrent"
+         wakeup from another task that did not clear
+         this entry from the heap before the
+         response process concluded. */
+      gc->suspended = false;
+      MHD_resume_connection (gc->connection);
+    }
+    GNUNET_assert (NULL != gc->hn);
+    gc->hn = NULL;
+    GNUNET_assert (gc ==
+                   GNUNET_CONTAINER_heap_remove_root (AH_to_heap));
+  }
+  if (NULL == gc)
+    return;
+  to_task = GNUNET_SCHEDULER_add_at (gc->timeout,
+                                     &do_timeout,
+                                     NULL);
+}
+
 
 void
 AH_truth_shutdown (void)
@@ -260,8 +316,8 @@ AH_truth_shutdown (void)
     }
     if (gc->suspended)
     {
-      MHD_resume_connection (gc->connection);
       gc->suspended = false;
+      MHD_resume_connection (gc->connection);
     }
     if (NULL != gc->as)
     {
@@ -271,6 +327,11 @@ AH_truth_shutdown (void)
     }
   }
   ANASTASIS_authorization_plugin_shutdown ();
+  if (NULL != to_task)
+  {
+    GNUNET_SCHEDULER_cancel (to_task);
+    to_task = NULL;
+  }
 }
 
 
@@ -401,6 +462,12 @@ request_done (struct TM_HandlerContext *hc)
                                  gc);
     gc->in_list = false;
   }
+  if (NULL != gc->hn)
+  {
+    GNUNET_assert (gc ==
+                   GNUNET_CONTAINER_heap_remove_node (gc->hn));
+    gc->hn = NULL;
+  }
   if (NULL != gc->as)
   {
     gc->authorization->cleanup (gc->as);
@@ -509,8 +576,8 @@ proposal_cb (void *cls,
                                gc);
   gc->in_list = false;
   GNUNET_assert (gc->suspended);
-  MHD_resume_connection (gc->connection);
   gc->suspended = false;
+  MHD_resume_connection (gc->connection);
   AH_trigger_daemon (NULL);
   if (MHD_HTTP_OK != por->hr.http_status)
   {
@@ -574,8 +641,8 @@ check_payment_cb (void *cls,
                                gc);
   gc->in_list = false;
   GNUNET_assert (gc->suspended);
-  MHD_resume_connection (gc->connection);
   gc->suspended = false;
+  MHD_resume_connection (gc->connection);
   AH_trigger_daemon (NULL);
 
   switch (hr->http_status)
@@ -828,7 +895,9 @@ run_authorization_process (struct MHD_Connection *connection,
   enum ANASTASIS_AUTHORIZATION_Result ret;
   enum GNUNET_DB_QueryStatus qs;
 
+  GNUNET_assert (! gc->suspended);
   ret = gc->authorization->process (gc->as,
+                                    gc->timeout,
                                     connection);
   switch (ret)
   {
@@ -853,7 +922,27 @@ run_authorization_process (struct MHD_Connection *connection,
     gc->as = NULL;
     return MHD_YES;
   case ANASTASIS_AUTHORIZATION_RES_SUSPENDED:
-    /* connection was suspended again, odd that this happens */
+    /* connection was suspended */
+    gc->suspended = true;
+    if (NULL == AH_to_heap)
+      AH_to_heap = GNUNET_CONTAINER_heap_create (
+        GNUNET_CONTAINER_HEAP_ORDER_MIN);
+    gc->hn = GNUNET_CONTAINER_heap_insert (AH_to_heap,
+                                           gc,
+                                           gc->timeout.abs_value_us);
+    if (NULL != to_task)
+    {
+      GNUNET_SCHEDULER_cancel (to_task);
+      to_task = NULL;
+    }
+    {
+      struct GetContext *rn;
+
+      rn = GNUNET_CONTAINER_heap_peek (AH_to_heap);
+      to_task = GNUNET_SCHEDULER_add_at (rn->timeout,
+                                         &do_timeout,
+                                         NULL);
+    }
     return MHD_YES;
   case ANASTASIS_AUTHORIZATION_RES_SUCCESS_REPLY_FAILED:
     /* Challenge sent successfully */
@@ -869,6 +958,20 @@ run_authorization_process (struct MHD_Connection *connection,
     gc->authorization->cleanup (gc->as);
     gc->as = NULL;
     return MHD_NO;
+  case ANASTASIS_AUTHORIZATION_RES_FINISHED:
+    GNUNET_assert (! gc->suspended);
+    gc->authorization->cleanup (gc->as);
+    gc->as = NULL;
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Resuming with authorization successful!\n");
+    if (gc->in_list)
+    {
+      GNUNET_CONTAINER_DLL_remove (gc_head,
+                                   gc_tail,
+                                   gc);
+      gc->in_list = false;
+    }
+    return MHD_YES;
   }
   GNUNET_break (0);
   return MHD_NO;
@@ -882,7 +985,6 @@ AH_handler_truth_get (
   struct TM_HandlerContext *hc)
 {
   struct GetContext *gc = hc->ctx;
-  struct GNUNET_HashCode challenge_response;
   void *encrypted_truth;
   size_t encrypted_truth_size;
   void *decrypted_truth;
@@ -967,7 +1069,7 @@ AH_handler_truth_get (
       if ( (NULL != challenge_response_s) &&
            (GNUNET_OK !=
             GNUNET_CRYPTO_hash_from_string (challenge_response_s,
-                                            &challenge_response)) )
+                                            &gc->challenge_response)) )
       {
         GNUNET_break_op (0);
         return TALER_MHD_reply_with_error (connection,
@@ -987,10 +1089,12 @@ AH_handler_truth_get (
       if (NULL != long_poll_timeout_ms)
       {
         unsigned int timeout;
+        char dummy;
 
         if (1 != sscanf (long_poll_timeout_ms,
-                         "%u",
-                         &timeout))
+                         "%u%c",
+                         &timeout,
+                         &dummy))
         {
           GNUNET_break_op (0);
           return TALER_MHD_reply_with_error (connection,
@@ -1013,6 +1117,10 @@ AH_handler_truth_get (
   } /* end of first-time initialization (if NULL == gc) */
   else
   {
+    /* might have been woken up by authorization plugin,
+       so clear the flag. MDH called us, so we are
+       clearly no longer suspended */
+    gc->suspended = false;
     if (NULL != gc->resp)
     {
       MHD_RESULT ret;
@@ -1032,6 +1140,7 @@ AH_handler_truth_get (
       GNUNET_assert (NULL != gc->authorization);
       GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                   "Continuing with running the authorization process\n");
+      GNUNET_assert (! gc->suspended);
       return run_authorization_process (connection,
                                         gc);
 
@@ -1074,8 +1183,8 @@ AH_handler_truth_get (
     {
       gc->authorization
         = ANASTASIS_authorization_plugin_load (method,
-                                               AH_cfg,
-                                               &gc->challenge_cost);
+                                               db,
+                                               AH_cfg);
       if (NULL == gc->authorization)
       {
         MHD_RESULT ret;
@@ -1090,6 +1199,7 @@ AH_handler_truth_get (
         GNUNET_free (method);
         return ret;
       }
+      gc->challenge_cost = gc->authorization->cost;
     }
     else
     {
@@ -1098,13 +1208,10 @@ AH_handler_truth_get (
     GNUNET_free (method);
   }
 
+  if ( (is_question) ||
+       (! gc->authorization->payment_plugin_managed) )
   {
-    struct TALER_Amount zero_amount;
-
-    TALER_amount_set_zero (AH_currency,
-                           &zero_amount);
-    if (0 != TALER_amount_cmp (&gc->challenge_cost,
-                               &zero_amount))
+    if (! TALER_amount_is_zero (&gc->challenge_cost))
     {
       /* Check database to see if the transaction is paid for */
       enum GNUNET_DB_QueryStatus qs;
@@ -1184,6 +1291,8 @@ AH_handler_truth_get (
      but check that the hash matches */
   if (is_question)
   {
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Handling security question challenge\n");
     if (! gc->have_response)
     {
       GNUNET_free (decrypted_truth);
@@ -1200,6 +1309,8 @@ AH_handler_truth_get (
       uint64_t code;
       enum ANASTASIS_DB_CodeStatus cs;
       struct GNUNET_HashCode hc;
+      bool satisfied;
+      uint64_t dummy;
 
       rt = GNUNET_TIME_UNIT_FOREVER_ABS;
       qs = db->create_challenge_code (db->cls,
@@ -1233,7 +1344,9 @@ AH_handler_truth_get (
                              &hc);
       cs = db->verify_challenge_code (db->cls,
                                       &gc->truth_uuid,
-                                      &hc);
+                                      &hc,
+                                      &dummy,
+                                      &satisfied);
       switch (cs)
       {
       case ANASTASIS_DB_CODE_STATUS_CHALLENGE_CODE_MISMATCH:
@@ -1256,15 +1369,15 @@ AH_handler_truth_get (
         GNUNET_assert (0);
       }
     }
-    if ( (decrypted_truth_size != sizeof (challenge_response)) ||
-         (0 != memcmp (&challenge_response,
+    if ( (decrypted_truth_size != sizeof (struct GNUNET_HashCode)) ||
+         (0 != memcmp (&gc->challenge_response,
                        decrypted_truth,
                        decrypted_truth_size)) )
     {
       GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                   "Wrong answer provided to secure question had %u bytes, wanted %u\n",
                   (unsigned int) decrypted_truth_size,
-                  (unsigned int) sizeof (challenge_response));
+                  (unsigned int) sizeof (struct GNUNET_HashCode));
       GNUNET_free (decrypted_truth);
       GNUNET_free (truth_mime);
       return TALER_MHD_reply_with_error (connection,
@@ -1282,17 +1395,21 @@ AH_handler_truth_get (
   if (gc->have_response)
   {
     enum ANASTASIS_DB_CodeStatus cs;
+    bool satisfied;
+    uint64_t code;
 
-    GNUNET_free (decrypted_truth);
     GNUNET_free (truth_mime);
     cs = db->verify_challenge_code (db->cls,
                                     &gc->truth_uuid,
-                                    &challenge_response);
+                                    &gc->challenge_response,
+                                    &code,
+                                    &satisfied);
     switch (cs)
     {
     case ANASTASIS_DB_CODE_STATUS_CHALLENGE_CODE_MISMATCH:
       GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                   "Provided response does not match our stored challenge\n");
+      GNUNET_free (decrypted_truth);
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_FORBIDDEN,
                                          TALER_EC_ANASTASIS_TRUTH_CHALLENGE_FAILED,
@@ -1300,98 +1417,111 @@ AH_handler_truth_get (
     case ANASTASIS_DB_CODE_STATUS_HARD_ERROR:
     case ANASTASIS_DB_CODE_STATUS_SOFT_ERROR:
       GNUNET_break (0);
+      GNUNET_free (decrypted_truth);
       return TALER_MHD_reply_with_error (gc->connection,
                                          MHD_HTTP_INTERNAL_SERVER_ERROR,
                                          TALER_EC_GENERIC_DB_FETCH_FAILED,
                                          "verify_challenge_code");
     case ANASTASIS_DB_CODE_STATUS_NO_RESULTS:
-      GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
-                  "No challenge known (challenge is invalidated after %u requests)\n",
-                  INITIAL_RETRY_COUNTER);
-      return TALER_MHD_reply_with_error (connection,
-                                         MHD_HTTP_TOO_MANY_REQUESTS,
-                                         TALER_EC_ANASTASIS_TRUTH_RATE_LIMITED,
-                                         NULL);
-    case ANASTASIS_DB_CODE_STATUS_VALID_CODE_STORED:
-      return return_key_share (&gc->truth_uuid,
-                               connection);
-    }
-    GNUNET_break (0);
-    return MHD_NO;
-  }
-
-  /* Not security question and no answer: use plugin to check if
-     decrypted truth is a valid challenge! */
-  {
-    enum GNUNET_GenericReturnValue ret;
-
-    ret = gc->authorization->validate (gc->authorization->cls,
-                                       connection,
-                                       truth_mime,
-                                       decrypted_truth,
-                                       decrypted_truth_size);
-    GNUNET_free (truth_mime);
-    switch (ret)
-    {
-    case GNUNET_OK:
-      /* data valid, continued below */
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                  "Response code unknown (possibly expired). Testing if we may provide a new one.\n");
+      gc->have_response = false;
       break;
-    case GNUNET_NO:
-      /* data invalid, reply was queued */
-      GNUNET_free (decrypted_truth);
-      return MHD_YES;
-    case GNUNET_SYSERR:
-      /* data invalid, reply was NOT queued */
-      GNUNET_free (decrypted_truth);
+    case ANASTASIS_DB_CODE_STATUS_VALID_CODE_STORED:
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                  "Response code valid (%s)\n",
+                  satisfied ? "satisfied" : "unsatisfied");
+      if (satisfied)
+      {
+        GNUNET_free (decrypted_truth);
+        return return_key_share (&gc->truth_uuid,
+                                 connection);
+      }
+      /* continue with authorization plugin below */
+      gc->code = code;
+      break;
+    default:
+      GNUNET_break (0);
       return MHD_NO;
     }
   }
-
-  /* Setup challenge and begin authorization process */
+  if (! gc->have_response)
   {
-    struct GNUNET_TIME_Absolute transmission_date;
-    enum GNUNET_DB_QueryStatus qs;
-
-    qs = db->create_challenge_code (db->cls,
-                                    &gc->truth_uuid,
-                                    gc->authorization->code_rotation_period,
-                                    gc->authorization->code_validity_period,
-                                    INITIAL_RETRY_COUNTER,
-                                    &transmission_date,
-                                    &gc->code);
-    switch (qs)
+    /* Not security question and no answer: use plugin to check if
+       decrypted truth is a valid challenge! */
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "No challenge provided, creating fresh challenge\n");
     {
-    case GNUNET_DB_STATUS_HARD_ERROR:
-    case GNUNET_DB_STATUS_SOFT_ERROR:
-      GNUNET_break (0);
-      GNUNET_free (decrypted_truth);
-      return TALER_MHD_reply_with_error (gc->connection,
-                                         MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                         TALER_EC_GENERIC_DB_FETCH_FAILED,
-                                         "store_challenge_code");
-    case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
-      /* 0 == retry_counter of existing challenge => rate limit exceeded */
-      GNUNET_free (decrypted_truth);
-      return TALER_MHD_reply_with_error (connection,
-                                         MHD_HTTP_TOO_MANY_REQUESTS,
-                                         TALER_EC_ANASTASIS_TRUTH_RATE_LIMITED,
-                                         NULL);
-    case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
-      /* challenge code was stored successfully*/
-      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-                  "Created fresh challenge\n");
-      break;
+      enum GNUNET_GenericReturnValue ret;
+
+      ret = gc->authorization->validate (gc->authorization->cls,
+                                         connection,
+                                         truth_mime,
+                                         decrypted_truth,
+                                         decrypted_truth_size);
+      GNUNET_free (truth_mime);
+      switch (ret)
+      {
+      case GNUNET_OK:
+        /* data valid, continued below */
+        break;
+      case GNUNET_NO:
+        /* data invalid, reply was queued */
+        GNUNET_free (decrypted_truth);
+        return MHD_YES;
+      case GNUNET_SYSERR:
+        /* data invalid, reply was NOT queued */
+        GNUNET_free (decrypted_truth);
+        return MHD_NO;
+      }
     }
 
-    if (GNUNET_TIME_absolute_get_duration (transmission_date).rel_value_us <
-        gc->authorization->code_retransmission_frequency.rel_value_us)
+    /* Setup challenge and begin authorization process */
     {
-      /* Too early for a retransmission! */
-      GNUNET_free (decrypted_truth);
-      return TALER_MHD_reply_with_error (gc->connection,
-                                         MHD_HTTP_ALREADY_REPORTED,
-                                         TALER_EC_ANASTASIS_TRUTH_CHALLENGE_ACTIVE,
-                                         NULL);
+      struct GNUNET_TIME_Absolute transmission_date;
+      enum GNUNET_DB_QueryStatus qs;
+
+      qs = db->create_challenge_code (db->cls,
+                                      &gc->truth_uuid,
+                                      gc->authorization->code_rotation_period,
+                                      gc->authorization->code_validity_period,
+                                      gc->authorization->retry_counter,
+                                      &transmission_date,
+                                      &gc->code);
+      switch (qs)
+      {
+      case GNUNET_DB_STATUS_HARD_ERROR:
+      case GNUNET_DB_STATUS_SOFT_ERROR:
+        GNUNET_break (0);
+        GNUNET_free (decrypted_truth);
+        return TALER_MHD_reply_with_error (gc->connection,
+                                           MHD_HTTP_INTERNAL_SERVER_ERROR,
+                                           TALER_EC_GENERIC_DB_FETCH_FAILED,
+                                           "create_challenge_code");
+      case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
+        /* 0 == retry_counter of existing challenge => rate limit exceeded */
+        GNUNET_free (decrypted_truth);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_TOO_MANY_REQUESTS,
+                                           TALER_EC_ANASTASIS_TRUTH_RATE_LIMITED,
+                                           NULL);
+      case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
+        /* challenge code was stored successfully*/
+        GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                    "Created fresh challenge\n");
+        break;
+      }
+
+      if (GNUNET_TIME_absolute_get_duration (transmission_date).rel_value_us <
+          gc->authorization->code_retransmission_frequency.rel_value_us)
+      {
+        /* Too early for a retransmission! */
+        GNUNET_free (decrypted_truth);
+        return TALER_MHD_reply_with_error (gc->connection,
+                                           MHD_HTTP_ALREADY_REPORTED,
+                                           TALER_EC_ANASTASIS_TRUTH_CHALLENGE_ACTIVE,
+                                           NULL);
+      }
     }
   }
 
@@ -1413,11 +1543,14 @@ AH_handler_truth_get (
                                        TALER_EC_ANASTASIS_TRUTH_AUTHORIZATION_START_FAILED,
                                        NULL);
   }
-  GNUNET_assert (! gc->in_list);
-  gc->in_list = true;
-  GNUNET_CONTAINER_DLL_insert (gc_head,
-                               gc_tail,
-                               gc);
+  if (! gc->in_list)
+  {
+    gc->in_list = true;
+    GNUNET_CONTAINER_DLL_insert (gc_head,
+                                 gc_tail,
+                                 gc);
+  }
+  GNUNET_assert (! gc->suspended);
   return run_authorization_process (connection,
                                     gc);
 }
