@@ -522,7 +522,7 @@ AH_handler_truth_post (
   struct ANASTASIS_CRYPTO_EncryptedKeyShareP keyshare_data;
   void *encrypted_truth;
   size_t encrypted_truth_size;
-  const char *truth_mime = "";
+  const char *truth_mime = NULL;
   const char *type;
   enum GNUNET_DB_QueryStatus qs;
   uint32_t storage_years;
@@ -594,10 +594,12 @@ AH_handler_truth_post (
       if (NULL != long_poll_timeout_ms)
       {
         unsigned int timeout;
+        char dummy;
 
         if (1 != sscanf (long_poll_timeout_ms,
-                         "%u",
-                         &timeout))
+                         "%u%c",
+                         &timeout,
+                         &dummy))
         {
           GNUNET_break_op (0);
           return TALER_MHD_reply_with_error (connection,
@@ -668,22 +670,18 @@ AH_handler_truth_post (
   }
 
   /* check method is supported */
+  if ( (0 != strcmp ("question",
+                     type)) &&
+       (NULL ==
+        ANASTASIS_authorization_plugin_load (type,
+                                             db,
+                                             AH_cfg)) )
   {
-    struct TALER_Amount dummy;
-
-    if ( (0 != strcmp ("question",
-                       type)) &&
-         (NULL ==
-          ANASTASIS_authorization_plugin_load (type,
-                                               AH_cfg,
-                                               &dummy)) )
-    {
-      GNUNET_JSON_parse_free (spec);
-      return TALER_MHD_reply_with_error (connection,
-                                         MHD_HTTP_BAD_REQUEST,
-                                         TALER_EC_ANASTASIS_TRUTH_UPLOAD_METHOD_NOT_SUPPORTED,
-                                         type);
-    }
+    GNUNET_JSON_parse_free (spec);
+    return TALER_MHD_reply_with_error (connection,
+                                       MHD_HTTP_BAD_REQUEST,
+                                       TALER_EC_ANASTASIS_TRUTH_UPLOAD_METHOD_NOT_SUPPORTED,
+                                       type);
   }
 
   if (storage_years > ANASTASIS_MAX_YEARS_STORAGE)
@@ -779,7 +777,9 @@ AH_handler_truth_post (
   qs = db->store_truth (db->cls,
                         truth_uuid,
                         &keyshare_data,
-                        truth_mime,
+                        (NULL == truth_mime)
+                        ? ""
+                        : truth_mime,
                         encrypted_truth,
                         encrypted_truth_size,
                         type,
