@@ -141,6 +141,23 @@ typedef void
   const struct TALER_Amount *amount);
 
 
+/**
+ * Function called to test if a given wire transfer
+ * satisfied the authentication requirement of the
+ * IBAN plugin.
+ *
+ * @param cls closure
+ * @param credit amount that was transferred
+ * @param wire_subject subject provided in the wire transfer
+ * @return true if this wire transfer satisfied the authentication check
+ */
+typedef bool
+(*ANASTASIS_DB_AuthIbanTransfercheck)(
+  void *cls,
+  const struct TALER_Amount *credit,
+  const char *wire_subject);
+
+
 /**
  * Handle to interact with the database.
  *
@@ -170,8 +187,26 @@ struct ANASTASIS_DatabasePlugin
    * @param cls closure
    * @return #GNUNET_OK upon success; #GNUNET_SYSERR upon failure
    */
-  int
-  (*drop_tables) (void *cls);
+  enum GNUNET_GenericReturnValue
+  (*drop_tables)(void *cls);
+
+  /**
+   * Connect to the database.
+   *
+   * @param cls closure
+   * @return #GNUNET_OK upon success; #GNUNET_SYSERR upon failure
+   */
+  enum GNUNET_GenericReturnValue
+  (*connect)(void *cls);
+
+  /**
+   * Initialize merchant tables
+   *
+   * @param cls closure
+   * @return #GNUNET_OK upon success; #GNUNET_SYSERR upon failure
+   */
+  enum GNUNET_GenericReturnValue
+  (*create_tables)(void *cls);
 
   /**
    * Function called to perform "garbage collection" on the
@@ -245,6 +280,7 @@ struct ANASTASIS_DatabasePlugin
    *
    * @param cls database context to use
    * @param es specification of the event to listen for
+   * @param timeout how long to wait for the event
    * @param cb function to call when the event happens, possibly
    *         multiple times (until cancel is invoked)
    * @param cb_cls closure for @a cb
@@ -253,6 +289,7 @@ struct ANASTASIS_DatabasePlugin
   struct GNUNET_DB_EventHandler *
   (*event_listen)(void *cls,
                   const struct GNUNET_DB_EventHeaderP *es,
+                  struct GNUNET_TIME_Relative timeout,
                   GNUNET_DB_EventCallback cb,
                   void *cb_cls);
 
@@ -567,13 +604,17 @@ struct ANASTASIS_DatabasePlugin
    * @param cls closure
    * @param truth_uuid identification of the challenge which the code corresponds to
    * @param hashed_code code which the user provided and wants to verify
+   * @param[out] code set to the original numeric code
+   * @param[out] satisfied set to true if the challenge is set to satisfied
    * @return transaction status
    */
   enum ANASTASIS_DB_CodeStatus
   (*verify_challenge_code)(
     void *cls,
     const struct ANASTASIS_CRYPTO_TruthUUIDP *truth_uuid,
-    const struct GNUNET_HashCode *hashed_code);
+    const struct GNUNET_HashCode *hashed_code,
+    uint64_t *code,
+    bool *satisfied);
 
 
   /**
@@ -585,7 +626,7 @@ struct ANASTASIS_DatabasePlugin
    * @param code code which is now satisfied
    * @return transaction status
    */
-  enum ANASTASIS_DB_CodeStatus
+  enum GNUNET_DB_QueryStatus
   (*mark_challenge_code_satisfied)(
     void *cls,
     const struct ANASTASIS_CRYPTO_TruthUUIDP *truth_uuid,
@@ -599,15 +640,17 @@ struct ANASTASIS_DatabasePlugin
    * @param cls closure
    * @param truth_uuid identification of the challenge which the code corresponds to
    * @param code code which is now satisfied
+   * @param after after what time must the challenge have been created
    * @return transaction status,
    *        #GNUNET_DB_STATUS_SUCCESS_NO_RESULTS if the challenge code is not satisfied or expired
    *        #GNUNET_DB_STATUS_SUCCESS_ONE_RESULT if the challenge code has been marked as satisfied
    */
-  enum ANASTASIS_DB_CodeStatus
+  enum GNUNET_DB_QueryStatus
   (*test_challenge_code_satisfied)(
     void *cls,
     const struct ANASTASIS_CRYPTO_TruthUUIDP *truth_uuid,
-    const uint64_t code);
+    const uint64_t code,
+    struct GNUNET_TIME_Absolute after);
 
 
   /**
@@ -632,7 +675,7 @@ struct ANASTASIS_DatabasePlugin
     const struct ANASTASIS_CRYPTO_TruthUUIDP *truth_uuid,
     struct GNUNET_TIME_Relative rotation_period,
     struct GNUNET_TIME_Relative validity_period,
-    unsigned int retry_counter,
+    uint32_t retry_counter,
     struct GNUNET_TIME_Absolute *retransmission_date,
     uint64_t *code);
 
@@ -714,6 +757,72 @@ struct ANASTASIS_DatabasePlugin
     const struct ANASTASIS_PaymentSecretP *payment_identifier);
 
 
+  /**
+   * Store inbound IBAN payment made for authentication.
+   *
+   * @param cls closure
+   * @param wire_reference unique identifier inside LibEuFin/Nexus
+   * @param wire_subject subject of the wire transfer
+   * @param amount how much was transferred
+   * @param debit_account account that was debited
+   * @param credit_account Anastasis operator account credited
+   * @param execution_date when was the transfer made
+   * @return transaction status
+   */
+  enum GNUNET_DB_QueryStatus
+  (*record_auth_iban_payment)(
+    void *cls,
+    uint64_t wire_reference,
+    const char *wire_subject,
+    const struct TALER_Amount *amount,
+    const char *debit_account,
+    const char *credit_account,
+    struct GNUNET_TIME_Absolute execution_date);
+
+
+  /**
+   * Function to check if we are aware of a wire transfer
+   * that satisfies the IBAN plugin's authentication check.
+   *
+   * @param cls closure
+   * @param debit_account which debit account to check
+   * @param earliest_date earliest date to check
+   * @param cb function to call on all entries found
+   * @param cb_cls closure for @a cb
+   * @return transaction status,
+   *    #GNUNET_DB_STATUS_SUCCESS_ONE_RESULT if @a cb
+   *      returned 'true' once
+   *    #GNUNET_DB_STATUS_SUCCESS_NO_RESULTS if no
+   *      wire transfers existed for which @a cb returned true
+   */
+  enum GNUNET_DB_QueryStatus
+  (*test_auth_iban_payment)(
+    void *cls,
+    const char *debit_account,
+    struct GNUNET_TIME_Absolute earliest_date,
+    ANASTASIS_DB_AuthIbanTransfercheck cb,
+    void *cb_cls);
+
+
+  /**
+   * Function to check the last known IBAN payment.
+   *
+   * @param cls closure
+   * @param credit_account which credit account to check
+   * @param[out] last_row set to the last known row
+   * @return transaction status,
+   *    #GNUNET_DB_STATUS_SUCCESS_ONE_RESULT if @a cb
+   *      returned 'true' once
+   *    #GNUNET_DB_STATUS_SUCCESS_NO_RESULTS if no
+   *      wire transfers existed for which @a cb returned true
+   */
+  enum GNUNET_DB_QueryStatus
+  (*get_last_auth_iban_payment_row)(
+    void *cls,
+    const char *credit_account,
+    uint64_t *last_row);
+
+
   /**
    * Function called to remove all expired codes from the database.
    *
