@@ -333,6 +333,34 @@ keyshare_lookup_cb (void *cls,
           = dd->details.server_failure.http_status
       };
 
+      c->af (c->af_cls,
+             &csr);
+      return;
+    }
+  case ANASTASIS_KSD_AUTHENTICATION_TIMEOUT:
+    {
+      struct ANASTASIS_ChallengeStartResponse csr = {
+        .cs = ANASTASIS_CHALLENGE_STATUS_AUTH_TIMEOUT,
+        .challenge = c,
+        .details.server_failure.ec
+          = dd->details.server_failure.ec,
+        .details.server_failure.http_status
+          = dd->details.server_failure.http_status
+      };
+
+      c->ci.async = true;
+      c->af (c->af_cls,
+             &csr);
+      return;
+    }
+  case ANASTASIS_KSD_EXTERNAL_CHALLENGE_INSTRUCTIONS:
+    {
+      struct ANASTASIS_ChallengeStartResponse csr = {
+        .cs = ANASTASIS_CHALLENGE_STATUS_EXTERNAL_INSTRUCTIONS,
+        .challenge = c,
+        .details.external_challenge = dd->details.external_challenge
+      };
+
       c->af (c->af_cls,
              &csr);
       return;
@@ -1009,7 +1037,9 @@ ANASTASIS_recovery_serialize (const struct ANASTASIS_Recovery *r)
       GNUNET_JSON_pack_string ("instructions",
                                c->instructions),
       GNUNET_JSON_pack_bool ("solved",
-                             c->ci.solved));
+                             c->ci.solved),
+      GNUNET_JSON_pack_bool ("async",
+                             c->ci.async));
     GNUNET_assert (0 ==
                    json_array_append_new (cs_arr,
                                           cs));
@@ -1084,6 +1114,9 @@ parse_cs_array (struct ANASTASIS_Recovery *r,
                                    &c->provider_salt),
       GNUNET_JSON_spec_string ("type",
                                &escrow_type),
+      GNUNET_JSON_spec_mark_optional (
+        GNUNET_JSON_spec_bool ("async",
+                               &c->ci.async)),
       GNUNET_JSON_spec_mark_optional (
         GNUNET_JSON_spec_fixed_auto ("key_share",
                                      &c->key_share)),
