@@ -151,6 +151,13 @@ struct SelectChallengeContext
    * Payment secret, if we are in the "pay" state.
    */
   struct ANASTASIS_PaymentSecretP ps;
+
+  /**
+   * Application asked us to only poll for existing
+   * asynchronous challenges, and not to being a
+   * new one.
+   */
+  bool poll_only;
 };
 
 
@@ -391,6 +398,51 @@ find_challenge_in_ri (json_t *state,
 }
 
 
+/**
+ * Find challenge of @a uuid in @a state under "cs".
+ *
+ * @param state the state to search
+ * @param uuid the UUID to search for
+ * @return NULL on error, otherwise challenge entry; RC is NOT incremented
+ */
+static json_t *
+find_challenge_in_cs (json_t *state,
+                      const struct ANASTASIS_CRYPTO_TruthUUIDP *uuid)
+{
+  json_t *rd = json_object_get (state,
+                                "recovery_document");
+  json_t *cs = json_object_get (rd,
+                                "cs");
+  json_t *c;
+  size_t off;
+
+  json_array_foreach (cs, off, c)
+  {
+    struct ANASTASIS_CRYPTO_TruthUUIDP u;
+    struct GNUNET_JSON_Specification spec[] = {
+      GNUNET_JSON_spec_fixed_auto ("uuid",
+                                   &u),
+      GNUNET_JSON_spec_end ()
+    };
+
+    if (GNUNET_OK !=
+        GNUNET_JSON_parse (c,
+                           spec,
+                           NULL, NULL))
+    {
+      GNUNET_break (0);
+      continue;
+    }
+    if (0 !=
+        GNUNET_memcmp (uuid,
+                       &u))
+      continue;
+    return c;
+  }
+  return NULL;
+}
+
+
 /**
  * Defines a callback for the response status for a challenge start
  * operation.
@@ -628,11 +680,10 @@ answer_feedback_cb (
                                           uuid,
                                           err));
     }
-    GNUNET_break_op (0);
     set_state (sctx->state,
                ANASTASIS_RECOVERY_STATE_CHALLENGE_SELECTING);
     sctx->cb (sctx->cb_cls,
-              TALER_EC_NONE,
+              csr->details.server_failure.ec,
               sctx->state);
     sctx_free (sctx);
     return;
@@ -653,7 +704,7 @@ answer_feedback_cb (
     set_state (sctx->state,
                ANASTASIS_RECOVERY_STATE_CHALLENGE_SELECTING);
     sctx->cb (sctx->cb_cls,
-              TALER_EC_NONE,
+              TALER_EC_ANASTASIS_TRUTH_UNKNOWN,
               sctx->state);
     sctx_free (sctx);
     return;
@@ -671,9 +722,120 @@ answer_feedback_cb (
                                           uuid,
                                           err));
     }
+    set_state (sctx->state,
+               ANASTASIS_RECOVERY_STATE_CHALLENGE_SELECTING);
+    sctx->cb (sctx->cb_cls,
+              TALER_EC_ANASTASIS_TRUTH_RATE_LIMITED,
+              sctx->state);
+    sctx_free (sctx);
+    return;
+  case ANASTASIS_CHALLENGE_STATUS_AUTH_TIMEOUT:
+    {
+      json_t *err;
+
+      err = GNUNET_JSON_PACK (
+        GNUNET_JSON_pack_string ("state",
+                                 "authentication-timeout"),
+        GNUNET_JSON_pack_uint64 ("error_code",
+                                 TALER_EC_ANASTASIS_TRUTH_AUTH_TIMEOUT));
+      GNUNET_assert (0 ==
+                     json_object_set_new (feedback,
+                                          uuid,
+                                          err));
+    }
     GNUNET_break_op (0);
     set_state (sctx->state,
                ANASTASIS_RECOVERY_STATE_CHALLENGE_SELECTING);
+    sctx->cb (sctx->cb_cls,
+              TALER_EC_ANASTASIS_TRUTH_AUTH_TIMEOUT,
+              sctx->state);
+    sctx_free (sctx);
+    return;
+
+  case ANASTASIS_CHALLENGE_STATUS_EXTERNAL_INSTRUCTIONS:
+    {
+      const json_t *body = csr->details.external_challenge;
+      const char *method;
+      json_t *details;
+      bool is_async = false;
+      uint64_t code = 0;
+      struct GNUNET_JSON_Specification spec[] = {
+        GNUNET_JSON_spec_string ("method",
+                                 &method),
+        GNUNET_JSON_spec_mark_optional (
+          GNUNET_JSON_spec_bool ("async",
+                                 &is_async)),
+        GNUNET_JSON_spec_mark_optional (
+          GNUNET_JSON_spec_uint64 ("answer_code",
+                                   &code)),
+        GNUNET_JSON_spec_json ("details",
+                               &details),
+        GNUNET_JSON_spec_end ()
+      };
+      json_t *reply;
+
+      if (GNUNET_OK !=
+          GNUNET_JSON_parse (body,
+                             spec,
+                             NULL, NULL))
+      {
+        json_t *err;
+
+        GNUNET_break_op (0);
+        err = GNUNET_JSON_PACK (
+          GNUNET_JSON_pack_string ("state",
+                                   "server-failure"),
+          GNUNET_JSON_pack_uint64 ("error_code",
+                                   TALER_EC_GENERIC_REPLY_MALFORMED));
+        GNUNET_assert (0 ==
+                       json_object_set_new (feedback,
+                                            uuid,
+                                            err));
+        return;
+      }
+      if (is_async)
+      {
+        json_t *c = find_challenge_in_cs (sctx->state,
+                                          &cd->uuid);
+
+        if (NULL == c)
+        {
+          GNUNET_break (0);
+          set_state (sctx->state,
+                     ANASTASIS_RECOVERY_STATE_ERROR);
+          sctx->cb (sctx->cb_cls,
+                    TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE,
+                    sctx->state);
+          sctx_free (sctx);
+          return;
+        }
+        GNUNET_assert (0 ==
+                       json_object_set_new (c,
+                                            "async",
+                                            json_true ()));
+        GNUNET_assert (0 ==
+                       json_object_set_new (c,
+                                            "answer-pin",
+                                            json_integer (code)));
+      }
+      reply = GNUNET_JSON_PACK (
+        GNUNET_JSON_pack_string ("state",
+                                 "external-instructions"),
+        GNUNET_JSON_pack_string ("method",
+                                 method),
+        GNUNET_JSON_pack_object_incref ("details",
+                                        details));
+      GNUNET_JSON_parse_free (spec);
+      GNUNET_assert (0 ==
+                     json_object_set_new (feedback,
+                                          uuid,
+                                          reply));
+    }
+    json_object_set_new (sctx->state,
+                         "selected_challenge_uuid",
+                         GNUNET_JSON_from_data_auto (&cd->uuid));
+    set_state (sctx->state,
+               ANASTASIS_RECOVERY_STATE_CHALLENGE_SOLVING);
     sctx->cb (sctx->cb_cls,
               TALER_EC_NONE,
               sctx->state);
@@ -719,7 +881,6 @@ solve_challenge_cb (void *cls,
                                  &ps),
     GNUNET_JSON_spec_end ()
   };
-
   json_t *challenge;
 
   if (NULL == ri)
@@ -733,10 +894,11 @@ solve_challenge_cb (void *cls,
     return;
   }
 
-  if (GNUNET_OK !=
-      GNUNET_JSON_parse (sctx->args,
-                         tspec,
-                         NULL, NULL))
+  if ( (NULL != sctx->args) &&
+       (GNUNET_OK !=
+        GNUNET_JSON_parse (sctx->args,
+                           tspec,
+                           NULL, NULL)) )
   {
     GNUNET_break_op (0);
     ANASTASIS_redux_fail_ (sctx->cb,
@@ -747,6 +909,80 @@ solve_challenge_cb (void *cls,
     return;
   }
 
+  /* resume all async, unsolved challenges */
+  {
+    bool poll_started = false;
+
+    for (unsigned int i = 0; i<ri->cs_len; i++)
+    {
+      struct ANASTASIS_Challenge *ci = ri->cs[i];
+      const struct ANASTASIS_ChallengeDetails *cd;
+      json_t *challenge;
+      json_t *pin;
+
+      cd = ANASTASIS_challenge_get_details (ci);
+      if (cd->solved ||
+          (! cd->async) )
+        continue;
+
+      challenge = find_challenge_in_cs (sctx->state,
+                                        &cd->uuid);
+      if (NULL == challenge)
+      {
+        GNUNET_break_op (0);
+        ANASTASIS_redux_fail_ (sctx->cb,
+                               sctx->cb_cls,
+                               TALER_EC_ANASTASIS_REDUCER_STATE_INVALID,
+                               "challenge not found");
+        sctx_free (sctx);
+        return;
+      }
+      pin = json_object_get (challenge,
+                             "answer-pin");
+      if (! json_is_integer (pin))
+      {
+        GNUNET_break_op (0);
+        ANASTASIS_redux_fail_ (sctx->cb,
+                               sctx->cb_cls,
+                               TALER_EC_ANASTASIS_REDUCER_STATE_INVALID,
+                               "async challenge 'answer-pin' not found");
+        sctx_free (sctx);
+        return;
+      }
+      if (GNUNET_OK !=
+          ANASTASIS_challenge_answer2 (ci,
+                                       psp,
+                                       timeout,
+                                       json_integer_value (pin),
+                                       &answer_feedback_cb,
+                                       sctx))
+      {
+        ANASTASIS_redux_fail_ (sctx->cb,
+                               sctx->cb_cls,
+                               TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE,
+                               "Failed to begin answering asynchronous challenge");
+        sctx_free (sctx);
+        return;
+      }
+      poll_started = true;
+    }
+
+    if (sctx->poll_only)
+    {
+      if (! poll_started)
+      {
+        GNUNET_break_op (0);
+        ANASTASIS_redux_fail_ (sctx->cb,
+                               sctx->cb_cls,
+                               TALER_EC_ANASTASIS_REDUCER_ACTION_INVALID,
+                               "no challenge available for polling");
+        return;
+      }
+      /* only polling, do not start new challenges */
+      return;
+    }
+  } /* end resuming async challenges */
+
   /* Check if we got a payment_secret */
   challenge = find_challenge_in_ri (sctx->state,
                                     &sctx->uuid);
@@ -801,13 +1037,17 @@ solve_challenge_cb (void *cls,
     psp = &ps;
   }
 
+  /* start or solve selected challenge */
   for (unsigned int i = 0; i<ri->cs_len; i++)
   {
     struct ANASTASIS_Challenge *ci = ri->cs[i];
     const struct ANASTASIS_ChallengeDetails *cd;
     int ret;
+    json_t *c;
 
     cd = ANASTASIS_challenge_get_details (ci);
+    if (cd->async)
+      continue; /* handled above */
     if (0 !=
         GNUNET_memcmp (&sctx->uuid,
                        &cd->uuid))
@@ -821,6 +1061,9 @@ solve_challenge_cb (void *cls,
       sctx_free (sctx);
       return;
     }
+    c = find_challenge_in_cs (sctx->state,
+                              &cd->uuid);
+    GNUNET_assert (NULL != c);
     if (0 == strcmp ("question",
                      cd->type))
     {
@@ -840,7 +1083,7 @@ solve_challenge_cb (void *cls,
       }
       /* persist answer, in case payment is required */
       GNUNET_assert (0 ==
-                     json_object_set (challenge,
+                     json_object_set (c,
                                       "answer",
                                       janswer));
       ret = ANASTASIS_challenge_answer (ci,
@@ -861,6 +1104,12 @@ solve_challenge_cb (void *cls,
       {
         uint64_t ianswer = json_integer_value (pin);
 
+        /* persist answer, in case async processing
+           happens via poll */
+        GNUNET_assert (0 ==
+                       json_object_set (c,
+                                        "answer-pin",
+                                        pin));
         ret = ANASTASIS_challenge_answer2 (ci,
                                            psp,
                                            timeout,
@@ -1130,6 +1379,67 @@ solve_challenge (json_t *state,
 }
 
 
+/**
+ * The user asked for us to poll on pending
+ * asynchronous challenges to see if they have
+ * now completed / been satisfied.
+ *
+ * @param[in] state we are in
+ * @param arguments our arguments with the solution
+ * @param cb functiont o call with the new state
+ * @param cb_cls closure for @a cb
+ * @return handle to cancel challenge selection step
+ */
+static struct ANASTASIS_ReduxAction *
+poll_challenges (json_t *state,
+                 const json_t *arguments,
+                 ANASTASIS_ActionCallback cb,
+                 void *cb_cls)
+{
+  struct SelectChallengeContext *sctx
+    = GNUNET_new (struct SelectChallengeContext);
+  json_t *rd;
+
+  rd = json_object_get (state,
+                        "recovery_document");
+  if (NULL == rd)
+  {
+    GNUNET_break_op (0);
+    ANASTASIS_redux_fail_ (cb,
+                           cb_cls,
+                           TALER_EC_ANASTASIS_REDUCER_STATE_INVALID,
+                           "poll_challenges");
+    return NULL;
+  }
+  sctx->poll_only = true;
+  sctx->cb = cb;
+  sctx->cb_cls = cb_cls;
+  sctx->state = json_incref (state);
+  sctx->args = json_incref ((json_t*) arguments);
+  sctx->r = ANASTASIS_recovery_deserialize (ANASTASIS_REDUX_ctx_,
+                                            rd,
+                                            &solve_challenge_cb,
+                                            sctx,
+                                            &core_secret_cb,
+                                            sctx);
+  if (NULL == sctx->r)
+  {
+    json_decref (sctx->state);
+    json_decref (sctx->args);
+    GNUNET_free (sctx);
+    GNUNET_break_op (0);
+    ANASTASIS_redux_fail_ (cb,
+                           cb_cls,
+                           TALER_EC_ANASTASIS_REDUCER_STATE_INVALID,
+                           "'recovery_document' invalid");
+    return NULL;
+  }
+  sctx->ra.cleanup = &sctx_free;
+  sctx->ra.cleanup_cls = sctx;
+  return &sctx->ra;
+}
+
+
 /**
  * The user selected a challenge to be solved. Handle the payment
  * process.
@@ -1387,12 +1697,33 @@ select_challenge_cb (void *cls,
       return;
     }
     /* trigger challenge */
-    ret = ANASTASIS_challenge_start (ci,
-                                     psp,
-                                     timeout,
-                                     NULL,   /* no answer */
-                                     &answer_feedback_cb,
-                                     sctx);
+    {
+      json_t *c = find_challenge_in_cs (sctx->state,
+                                        &cd->uuid);
+      json_t *pin = json_object_get (c,
+                                     "answer-pin");
+
+      if (NULL != pin)
+      {
+        uint64_t ianswer = json_integer_value (pin);
+
+        ret = ANASTASIS_challenge_answer2 (ci,
+                                           psp,
+                                           timeout,
+                                           ianswer,
+                                           &answer_feedback_cb,
+                                           sctx);
+      }
+      else
+      {
+        ret = ANASTASIS_challenge_start (ci,
+                                         psp,
+                                         timeout,
+                                         NULL, /* no answer */
+                                         &answer_feedback_cb,
+                                         sctx);
+      }
+    }
     if (GNUNET_OK != ret)
     {
       ANASTASIS_redux_fail_ (sctx->cb,
@@ -1688,6 +2019,11 @@ ANASTASIS_recovery_action_ (json_t *state,
       "select_challenge",
       &select_challenge
     },
+    {
+      ANASTASIS_RECOVERY_STATE_CHALLENGE_SELECTING,
+      "poll",
+      &poll_challenges
+    },
     {
       ANASTASIS_RECOVERY_STATE_CHALLENGE_SELECTING,
       "back",
@@ -2283,6 +2619,10 @@ launch_recovery (struct RecoverSecretState *rss,
     GNUNET_JSON_spec_end ()
   };
 
+  if (MHD_HTTP_OK !=
+      json_integer_value (json_object_get (p_cfg,
+                                           "http_status")))
+    return false; /* skip providers that are down */
   if (GNUNET_OK !=
       GNUNET_JSON_parse (p_cfg,
                          spec,
