@@ -570,8 +570,13 @@ config_cb (void *cls,
     }
     else
     {
+      GNUNET_free (cr->currency);
       cr->currency = GNUNET_strdup (acfg->currency);
+      GNUNET_free (cr->business_name);
       cr->business_name = GNUNET_strdup (acfg->business_name);
+      for (unsigned int i = 0; i<cr->methods_length; i++)
+        GNUNET_free (cr->methods[i].type);
+      GNUNET_free (cr->methods);
       cr->methods = GNUNET_new_array (acfg->methods_length,
                                       struct AuthorizationMethodConfig);
       for (unsigned int i = 0; i<acfg->methods_length; i++)
